# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>. 1

"""Master module concerning readme plugins."""

import glob
import os

from evariste import errors, plugins

class CannotRender(errors.EvaristeError):
    """Exception raised when renderer cannot render a given file."""
    pass

class ReadmeRenderer(plugins.PluginBase):
    """Generic class of README renderers, and also a default raw renderer."""
    # pylint: disable=too-few-public-methods, abstract-method

    extensions = []

    plugin_type = ["renderer", "html", "readme"]

    def __init__(self, setup):
        super().__init__(setup)
        # TODO Make it read from the setup file the list of extensions to
        # render as raw.

    def render(self, path, *args, **kwargs):
        """Render the readme file ``path``.
        """
        # pylint: disable=unused-argument, no-self-use
        with open(path) as file:
            return file.read()

    def iter_readme(self, value, directory):
        """Iterave over potential readme files for the path ``value``."""
        if directory:
            yield from self.iter_readme_dir(value)
        else:
            yield from self.iter_readme_file(value)

    def iter_readme_dir(self, value):
        """Iterate over potential readme for the given directory ``value``."""
        for ext in self.extensions:
            for filename in glob.iglob((value / "*.{}".format(ext)).as_posix()):
                basename = os.path.basename(filename)
                if basename.count(".") == 1:
                    if basename.split(".")[0].lower() == "readme":
                        yield filename

    def iter_readme_file(self, value):
        """Iterate over potential readme for the given file ``value``."""
        for ext in self.extensions:
            for filename in glob.iglob("{}.{}".format(value, ext)):
                yield filename




# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>. 1

"""A rag-bag of utility functions that did not fit anywhere else."""

import os
import shutil

def copy(source, destination):
    """Copy ``source`` to ``destination``, creating directories if necessary."""
    os.makedirs(os.path.dirname(destination), exist_ok=True)
    shutil.copy(source, destination)

### Partial format
### TODO There must be a cleaner way to do this. I have no Internet right now
def partial_format(string, *args, **kwargs):
    """Format string, leaving unfilled fields untouched.

    >>> partial_format("To {verb} or {negation} to {verb}", negation="not")
    "To {verb} or not to {verb}"
    """

    def fields(string):
        """Iterator over the fields of the string.

        >>> list(fields("To {be} or {not} to be"))
        ["be", "not"]
        """
        store = False
        current = ""
        for char in string:
            if char == '{':
                store = True
            elif char == '}':
                store = False
                yield current
                current = ""
            else:
                if store:
                    current += char

    dictionary = {}
    for field in fields(string):
        dictionary[field] = '{' + field + '}'

    dictionary.update(kwargs)
    return string.format(*args, **dictionary) # pylint: disable=star-args



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>. 1

"""Command line client to :mod:`evariste`"""

import logging
import sys

from evariste import errors, tree, renderer
from evariste.action import load_actions
from evariste.main.options import get_options
from evariste.rcs import git
from evariste.setup import Setup
import evariste

LOGGER = logging.getLogger(evariste.__name__)
LOGGER.addHandler(logging.StreamHandler())
logging.addLevelName(15, "DETAIL")

def main():
    """Main function"""
    arguments = get_options()
    try:
        LOGGER.info("Reading setup…")
        setup = Setup.from_file(arguments.setup) # pylint: disable=no-member
        LOGGER.info("Building directory tree…")
        directory = tree.build_tree(
            git.Git(setup['setup']['source']),
            setup,
            )
        LOGGER.info("Compiling…")
        directory.compile_root(load_actions(setup))

        renderers = evariste.renderer.load_renderers(setup)
        for keyword in setup["renderer"]["plugins"]:
            LOGGER.info("Rendering {}…".format(keyword))
            print(renderers.match(keyword).render(directory))
    except errors.EvaristeError as error:
        LOGGER.error("Error: " + str(error))
        sys.exit(1)



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>. 1

"""Compilation of {,La,Lua,Xe}TeX files"""

import os

from evariste.action.command import MultiCommand
from evariste import utils

class CompilationEnvironment:
    """Environment of compilation.

    Stores some variables related to the compilation of one LaTeX file.
    """
    # pylint: disable=too-few-public-methods

    BIN = {
        'latex': 'latex',
        'bibtex': 'bibtex',
        # INDEX
        }

    def __init__(self, path):
        self.bin = self.BIN.copy()
        self.path = path
        self.latex_todo = 1

class Latex(MultiCommand):
    """Compilation of *TeX files."""

    # TODO add index generation (look for INDEX in this file)

    keyword = "latex"
    priority = 100

    DEFAULT_COMMANDS = [
        "latex",
        "?bibtex",
        # INDEX
        "?latex",
        ]
    PREDEFINED_COMMANDS = {
        'latex': '{latex} {basename}',
        'pdflatex': 'pdflatex {basename}',
        'lualatex': 'lualatex {basename}',
        'xelatex': 'xelatex {basename}',
        'bibtex': '{bibtex} {basename}',
        'biblatex': 'biblatex {basename}',
        # INDEX
        }
    TARGET = {
        'latex': '{basename}.dvi',
        'pdflatex': '{basename}.pdf',
        'lualatex': '{basename}.TODO', # TODO
        'xelatex': '{basename}.TODO', # TODO
        }
    CONDITIONAL_COMMANDS = [
        'latex',
        'bibtex',
        # INDEX
        ]

    def match(self, value):
        return value.suffix == ".tex"

    def _compile(self, path):
        # pylint: disable=arguments-differ
        return super()._compile(path, CompilationEnvironment(path))

    def _read_config(self, path, env):
        # pylint: disable=signature-differs
        # Reading commands
        precommands = []
        postcommands = []
        commands = []
        if self.setup_section in env.path.config:
            for option in env.path.config[self.setup_section]:
                if option.startswith("pre"):
                    precommands.append("!{}".format(env.path.config[self.setup_section][option]))
                elif option.startswith("cmd"):
                    commands.append(env.path.config[self.setup_section][option])
                elif option.startswith("post"):
                    postcommands.append("!{}".format(env.path.config[self.setup_section][option]))
        if not commands:
            commands = self.DEFAULT_COMMANDS
        self._commands = precommands + commands + postcommands

        # Reading binaries
        for key in env.bin:
            if env.path.config.has_option(self.setup_section, key):
                if env.path.config.get(self.setup_section, key) is not None:
                    env.bin[key] = env.path.config.get(self.setup_section, key)

        # Setting target
        if not env.path.config.has_option(self.setup_section, 'target'):
            basetarget = env.path.format(self.TARGET[env.bin['latex']])
        else:
            basetarget = env.path.format(env.path.config[self.setup_section]["target"])
        self._report.target = os.path.join(
            env.path.parent.from_source.as_posix(),
            basetarget,
            )

    def _format_command(self, command, env):
        """Iterate over commands that are to be executed."""
        if command.startswith("?") and command[1:] in self.CONDITIONAL_COMMANDS:
            yield from getattr(self, "_conditionnal_command_{}".format(command[1:]))(env)
        elif command.startswith("!"):
            yield utils.partial_format(command[1:], **env.bin)
        elif command in self.PREDEFINED_COMMANDS:
            yield utils.partial_format(self.PREDEFINED_COMMANDS[command], **env.bin)
        else:
            yield utils.partial_format(command, **env.bin)

    def iter_commands(self, env):
        for command in self._commands:
            yield from self._format_command(command, env)

    def _conditionnal_command_bibtex(self, env):
        """Yield a bibtex command, if necessary.

        "Necessary" means that option 'bibtex' appears in configuration file.

        It might be detected, some day, by looking at the LaTeX log.
        """
        if env.path.config.has_option(self.setup_section, 'bibtex'):
            env.latex_todo += 1
            yield from self._format_command('bibtex', env)

    def _conditionnal_command_latex(self, env):
        """Yield as many `latex` commands as necessary.

        Previous commands may have incremented `env.latex_todo`.
        """
        while env.latex_todo != 0:
            env.latex_todo -= 1
            yield from self._format_command('latex', env)



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>. 1

"""TODO Parse setup file."""

import collections
import configparser

_DEFAULT_SETUP = {
    'setup': {
        'source': '.',
        },
    'renderer': {
        'plugins': '',
        },
    }

def Options(*args, **kwargs):
    """Return a :class:`collections.defaultdict` object, default being `None`.
    """
    # pylint: disable=invalid-name
    return collections.defaultdict(lambda: None, *args, **kwargs)

def Sections(*args, **kwargs):
    """Return a :class:`collections.defaultdict` object, default being `Options`
    """
    # pylint: disable=invalid-name
    return collections.defaultdict(Options, *args, **kwargs)

class Setup:
    """Representation ef Evariste setup."""

    def __init__(self, dictionary=None):
        if dictionary is None:
            dictionary = {}
        self.dict = Sections()
        self.fill_blanks(dictionary)
        self.fill_blanks(_DEFAULT_SETUP)

        for section in ["renderer", "action"]:
            if "plugins" in self[section]:
                self[section]["plugins"] = self[section]["plugins"].strip().split()

    def __iter__(self):
        yield from self.dict

    def __getitem__(self, value):
        return self.dict[value]

    def __str__(self):
        return "{{{}}}".format(", ".join([
            "{}: {}".format(key, value)
            for key, value
            in self.dict.items()
            ]))

    def fill_blanks(self, dictionary):
        """Fill unset self options with argument."""
        for section in dictionary:
            if section in self.dict:
                for option in dictionary[section]:
                    if option not in self.dict[section]:
                        self.dict[section][option] = dictionary[section][option]
            else:
                self.dict[section] = Options(dictionary[section])

    @classmethod
    def from_file(cls, filename):
        """Parse configuration file ``filename``."""
        with open(filename) as file:
            return cls.from_string(file.read())

    @classmethod
    def from_string(cls, string):
        """Parse ``string`` as the content of a configuration file."""
        config = configparser.ConfigParser(allow_no_value=True)
        config.read_string(string)
        return cls.from_config(config)

    @classmethod
    def from_config(cls, setup):
        """Parse ``setup`` as a :class:`configparser.ConfigParser` object."""
        dictionary = dict()
        for section in setup:
            if section == "DEFAULT":
                continue
            dictionary[section] = dict()
            for option in setup[section]:
                dictionary[section][option] = setup[section][option]
        return cls(dictionary)



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>. 1

"""Access to RCS (git, etc.) versionned files."""

import os
import pathlib

from evariste import errors
from evariste.tree import Directory

class NoRepositoryError(errors.EvaristeError):
    """No repository contains the given path."""

    def __init__(self, rcstype, directory):
        super().__init__()
        self.directory = directory
        self.rcstype = rcstype

    def __str__(self):
        return (
            "Could not find any {} repository containing directory '{}'."
            ).format(
                self.rcstype,
                self.directory,
                )

class RCS:
    """Generic class to access to versionned files."""

    def __init__(self, directoryname):
        self.from_source = pathlib.Path(directoryname)

    def tree(self, setup):
        """Return a :class:`Directory` object representing the versionned files.
        """
        tree = Directory(self.from_source, rcs=self, setup=setup)
        for path in self.walk():
            tree.add_subpath(pathlib.Path(path))
        return tree

    def walk(self):
        """Iterate over list of versionned files, descendants of ``self.from_source``.
        """
        raise NotImplementedError()

    def is_versionned(self, path):
        """Return ``True`` iff ``path`` is versionned."""
        raise NotImplementedError()

    @property
    def workdir(self):
        """Return path of the root of the repository."""
        raise NotImplementedError()

    def from_repo(self, path):
        """Return ``path``, relative to the repository root."""
        raise NotImplementedError()

    def last_modified(self, path):
        """Return the datetime of last modification."""
        raise NotImplementedError()



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>. 1

"""Target renderers: how should target files be rendered?"""

from evariste import errors, plugins

class TargetRenderer(plugins.PluginBase):
    """Renderer of target (compiled) files."""

    def match(self, dummy):
        """This is the default renderer, that matches everything."""
        return True

    def render(self, tree):
        """Render ``tree``, which is a :class:`tree.File`."""
        raise NotImplementedError()



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>. 1

"""Render tree as an HTML (body) page."""

import logging

from evariste.renderer.jinja2 import JinjaRenderer
from evariste.renderer.html import target, readme
import evariste

LOGGER = logging.getLogger(evariste.__name__)

class HTMLRenderer(JinjaRenderer):
    """Render tree as an HTML div (without the `<div>` tags)."""
    # pylint: disable=too-few-public-methods

    subplugin_baseclasses = {
        "target": target.HtmlRenderer,
        "readme": readme.HtmlRenderer,
        }
    keyword = "html"
    extension = "html"
    description = "Render as an html div"
    default_setup = {
        "href_prefix": "",
        }



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>. 1

"""Access to git-versionned files."""

from datetime import datetime
import itertools
import os
import pygit2

from evariste import rcs

RCS_TYPE = 'git'

class Git(rcs.RCS):
    """Access git-versionned files"""
    # pylint: disable=no-member

    def __init__(self, directoryname):
        super().__init__(directoryname)
        try:
            self.repository = pygit2.Repository(
                pygit2.discover_repository(
                    directoryname
                    )
                )
        except KeyError:
            raise rcs.NoRepositoryError(RCS_TYPE, directoryname)

    def walk(self):
        root = self.from_source.resolve().relative_to(self.workdir)
        for entry in self.repository.index:
            if entry.path.startswith(root.as_posix()):
                yield os.path.relpath(entry.path, root.as_posix())

    def is_versionned(self, path):
        from_repo = self.from_repo(path)
        for entry in self.repository.index:
            if from_repo == entry.path:
                return True
        return False

    @property
    def workdir(self):
        return self.repository.workdir

    def from_repo(self, path):
        return os.path.relpath(
            path.resolve().as_posix(),
            self.workdir
            )

    def _walk_for_files(self, path):
        """TODO

        Copied and adapted from
        https://github.com/libgit2/pygit2/issues/200#issuecomment-15899713
        """
        walker = self.repository.walk(
            self.repository.head.target,
            pygit2.GIT_SORT_TIME,
            )
        commit0 = next(walker)
        for commit1 in walker:
            files_in_diff = (
                (x.old_file_path, x.new_file_path)
                for x in commit1.tree.diff_to_tree(commit0.tree)
                )
            if path in itertools.chain(*files_in_diff): # pylint: disable=star-args
                return commit0
            commit0 = commit1

    def last_modified(self, path):
        return datetime.fromtimestamp(self._walk_for_files(self.from_repo(path)).commit_time)




# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>. 1

"""Abstract class for jinja2 renderers."""

import jinja2
import os
import pkg_resources
import pathlib
import textwrap

from evariste import utils
from evariste.renderer.dest import DestRenderer
from evariste.target.jinja2 import JinjaRenderer as TargetJinjaRenderer
from evariste.readme.jinja2 import JinjaRenderer as ReadmeJinjaRenderer

class JinjaRenderer(DestRenderer):
    """Abstract class for jinja2 renderers."""
    # pylint: disable=too-few-public-methods

    extension = None
    subplugin_baseclasses = {
        "target": TargetJinjaRenderer,
        "readme": ReadmeJinjaRenderer,
        }

    def __init__(self, setup):
        super().__init__(setup)

        self.environment = jinja2.Environment(
            loader=jinja2.FileSystemLoader(
                os.path.dirname(pkg_resources.resource_filename( #pylint: disable=no-member
                    self.__class__.__module__,
                    os.path.join("data", "templates", "tree.{}".format(self.extension))
                ))
            )
        )
        self.environment.filters['basename'] = os.path.basename


    def render(self, tree):
        return self._render(
            tree=tree,
            template="tree.{}".format(self.extension),
            context={
                'destdir': pathlib.Path(self.destdir),
                'setup': self.setup,
                'global_setup': self.global_setup,
                'render': self._render,
                'render_download': self._render_download,
                'render_readme': self._render_readme,
                'render_file': self._render_file,
                'render_directory': self._render_directory,
                'render_template': self._render_template,
                },
            )

    @jinja2.contextfunction
    def _render(self, context, tree, template):
        """Render the tree, using given template (or template list)."""
        if isinstance(context, dict):
            context['tree'] = tree
        else:
            context.vars['tree'] = tree
        if tree.is_file() and tree.report.success:
            utils.copy(
                os.path.join(
                    tree.root.from_cwd.as_posix(),
                    tree.report.target,
                    ),
                os.path.join(
                    context['destdir'].as_posix(),
                    tree.report.target,
                    )
                )
        return textwrap.indent(
            self.environment.get_or_select_template(template).render(context),
            "  ",
            )

    @jinja2.contextfunction
    def _render_directory(self, context, tree):
        """Render ``tree``, which is a :class:`tree.Directory`, as HTML."""
        return self._render(context, tree, "tree_directory.{}".format(self.extension))

    @jinja2.contextfunction
    def _render_file(self, context, tree):
        """Render ``tree``, which is a :class:`tree.File`."""
        return self.subplugins["target"].match(tree).render(tree, context)

    @jinja2.contextfunction
    def _render_download(self, context, tree):
        """Render the code downloading the archive."""
        path = tree.make_archive(context["destdir"])
        context.vars['href'] = path.as_posix()
        context.vars['content'] = path.name
        return self.environment.get_template("download.{}".format(self.extension)).render(context) # pylint: disable=no-member

    @jinja2.contextfunction
    def _render_readme(self, context, tree):
        """Find the readme of tree, and returns the corresponding code."""
        # pylint: disable=unused-argument
        return tree.render_readme(self.subplugins["readme"], context)

    @jinja2.contextfunction
    def _render_template(self, context, template):
        """Render template given in argument."""
        with open(template) as file:
            return self.environment.from_string(file.read()).render(context)



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>. 1

"""Render (processed) tree: abstract class."""

import jinja2
import os
import pathlib
import pkg_resources
import tempfile
import textwrap

from evariste import errors, plugins, readme, target

def load_renderers(setup):
    """Return the :class:`PluginLoader`` of renderer plugins."""
    return plugins.PluginLoader(["renderer"], Renderer, setup)

class Renderer(plugins.PluginBase):
    """Abstract tree renderer."""

    plugin_type = ["renderer"]

    plugin_baseclasses = {
        "readme": readme.ReadmeRenderer,
        "target": target.TargetRenderer,
        }

    def priority(self):
        """Return keyword as priority.

        That way, plugins are sorted by name.
        """
        return self.keyword

    def match(self, value):
        """Return ``True`` iff ``value`` is the keyword."""
        return value == self.keyword

    def render(self, tree):
        """Render tree."""
        raise NotImplementedError()




# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>. 1

"""Raw action: does nothing; simply provides a link to download file."""

from evariste import action


class Raw(action.Action):
    """Raw action"""
    # pylint: disable=too-few-public-methods

    keyword = "raw"
    priority = -100

    # TODO complete:
    # - archives
    # - libreoffice / Ms Office
    # See if it is not better with mimetypes.
    MATCH_SUFFIXES = [
        ".pdf", ".dvi",
        ".png", ".jpg", ".jpeg", ".gif",
        ]

    def _compile(self, path):
        return action.Report(
            path,
            target=path.from_source.as_posix(),
            success=True,
            )

    def match(self, value):
        return value.suffix in self.MATCH_SUFFIXES



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>. 1


"""Plugin loader"""

from functools import total_ordering
import importlib
import logging
import os
import pkgutil
import sys

from evariste import errors
import evariste

LOGGER = logging.getLogger(evariste.__name__)

class NoMatch(errors.EvaristeError):
    """No plugin found matching ``value``."""

    def __init__(self, value, available):
        super().__init__()
        self.value = value
        self.available = available

    def __str__(self):
        return "Value '{}' does not match any of {}.".format(
            self.value,
            str(self.available),
            )

class SameKeyword(errors.EvaristeError):
    """Two plugins have the same keyword."""

    def __init__(self, base, keyword, plugin1, plugin2):
        super().__init__()
        self.base = base
        self.keyword = keyword
        self.plugins = (plugin1, plugin2)

    def __str__(self):
        return """Plugins '{}' and '{}' (from '{}') have the same keyword '{}'.""".format(
            self.plugins[0].__name__,
            self.plugins[1].__name__,
            self.base,
            self.keyword,
            )

class NotAPlugin(errors.EvaristeError):
    """Superclass of plugins is not a plugin."""

    def __init__(self, obj):
        super().__init__()
        self.obj = obj

    def __str__(self):
        return (
            """Class '{obj.__module__}.{obj.__name__}' is not a plugin """
            """(it should inherit from """
            """'{superclass.__module__}.{superclass.__name__}')."""
            ).format(
                obj=self.obj,
                superclass=PluginBase,
            )

class PluginBase:
    """Plugin base: all imported plugins must be subclasses of this class."""
    # pylint: disable=too-few-public-methods
    keyword = None
    description = None
    priority = 0
    default_setup = {}
    plugin_type = []
    subplugin_baseclasses = {}
    subplugins = {}

    def __init__(self, setup):
        self.global_setup = setup
        self._set_default_setup()
        self._load_plugins()

    def _set_default_setup(self):
        """Set default value for this plugin setup, if necessary."""
        for parent in self.__class__.mro(): # pylint: disable=no-member
            if hasattr(parent, "default_setup"):
                self.global_setup.fill_blanks(self.default_setup)

    def _load_plugins(self):
        """Load subplugins."""
        for (keyword, baseclass) in self.subplugin_baseclasses.items():
            self.subplugins[keyword] = PluginLoader(
                self.__module__.split('.')[1:] + [keyword],
                baseclass,
                self.global_setup,
                )

    @property
    def setup(self):
        """Return this plugin setup."""
        return self.global_setup[self.setup_section]

    def match(self, value, *args, **kwargs): # pylint: disable=unused-argument
        """Return ``True`` iff ``value`` is to be compiled by ``self``."""
        raise NotImplementedError()

    @property
    def setup_section(self):
        """Return the name of the section of the object, in the setup file."""
        return ".".join(self.plugin_type + [self.keyword])

@total_ordering
class PluginEntry:
    """Plugin, with associated keyword and priority."""
    # pylint: disable=too-few-public-methods

    def __init__(self, plugin):
        self.plugin = plugin
        self.keyword = plugin.keyword
        self.priority = plugin.priority

    def match(self, value, *args, **kwargs):
        """Call :meth:`self.plugin.match`."""
        return self.plugin.match(value, *args, **kwargs)

    def __eq__(self, other):
        return self.plugin == other.plugin

    def __lt__(self, other):
        priority = self.priority
        if callable(priority):
            priority = priority()
        other_priority = other.priority
        if callable(other_priority):
            other_priority = other_priority()
        if priority == other_priority:
            return self.keyword < other.keyword
        return priority < other_priority

    def __hash__(self):
        return hash(self.plugin)

def expand_path(path):
    """Return ``path`` where environment variables and user directory have been expanded."""
    return os.path.expanduser(os.path.expandvars(path))

class PluginLoader:
    """Plugin loader. Find plugins and select the appropriate one."""

    def __init__(self, base, superclass, setup, libdirs=None):
        """Initialization.

        :param list base: Base module. If ``base == ["foo", "bar"]``, plugins
            are supposed to be submodules of ``evariste.foo.bar``, in
            directory ``evariste/foo/bar``.
        :param type superclass: Base class. Every subclass of ``superclass``
            is a plugin.
        """
        if not issubclass(superclass, PluginBase):
            raise NotAPlugin(superclass)

        self.plugins = set()
        if libdirs is None:
            libdirs = []
        base = ["evariste"] + base

        path = []
        path.extend([
            expand_path(item)
            for item
            in [".evariste", "~/.config/evariste", "~/.evariste"]
            ])
        path.extend([expand_path(item) for item in libdirs])
        path.extend(sys.path)
        evariste_path = [
            os.path.join(item, *base) # pylint: disable=star-args
            for item
            in path
            ]

        for module_finder, name, is_pkg in pkgutil.iter_modules(
                evariste_path,
                prefix=".".join(base) + "."
            ):
            try:
                module = module_finder.find_spec(name).loader.load_module()
            except ImportError:
                continue
            for attr in dir(module):
                if attr.startswith("_"):
                    continue
                obj = getattr(module, attr)

                if (
                        isinstance(obj, type)
                        and
                        issubclass(obj, superclass)
                    ):
                    if obj.keyword is None:
                        continue
                    if obj.keyword in self:
                        if obj != self[obj.keyword].plugin.__class__:
                            raise SameKeyword(
                                "{}.{}".format(
                                    superclass.__module__,
                                    superclass.__name__,
                                    ),
                                obj.keyword,
                                obj,
                                self[obj.keyword].plugin.__class__,
                                )
                    self.plugins.add(PluginEntry(obj(setup)))

    def match(self, value):
        """Return the first plugin matching ``value``.

        A plugin ``Foo`` matches ``value`` if ``Foo.match(value)`` returns
        True.
        """
        for entry in sorted(self, reverse=True):
            if entry.match(value):
                return entry.plugin
        raise NoMatch(value, sorted(self.iter_keywords()))

    def iter_plugins(self):
        """Iterate over plugins (as classes)."""
        for entry in self:
            yield entry.plugin

    def __iter__(self):
        for plugin in self.plugins:
            yield plugin

    def __contains__(self, keyword):
        return keyword in self.iter_keywords()

    def iter_keywords(self):
        """Iterate over plugin keywords."""
        for plugin in self:
            yield plugin.keyword

    def __getitem__(self, keyword):
        for entry in self:
            if entry.keyword == keyword:
                return entry
        raise KeyError(keyword)



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>. 1

"""Common utilities for readme renderers using Jinja2"""

import textwrap

from evariste import readme

class JinjaRenderer(readme.ReadmeRenderer):
    """Abstract class for readme renderers using jinja2."""
    # pylint: disable=too-few-public-methods, abstract-method

    extensions = []

    def render(self, path, context):
        """Render file as html code."""
        # pylint: disable=arguments-differ
        return textwrap.indent(
            context['render_template'](context, path),
            "  ",
            )



#!/usr/bin/env python3

# Copyright 2015 Louis Paternault
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Installer"""

from setuptools import setup, find_packages
import codecs
import os

def readme():
    directory = os.path.dirname(os.path.join(
        os.getcwd(),
        __file__,
        ))
    return codecs.open(os.path.join(directory, "README.rst"), "r", "utf8").read()

setup(
        name='Evariste',
        version="0.0.0",
        packages=find_packages(),
        setup_requires=["hgtools"],
        install_requires=[
            "jinja2",
            "docutils",
            "pygit2",
            ],
        include_package_data=True,
        author='Louis Paternault',
        author_email='spalax@gresille.org',
        description='TODO',
        url='https://git.framasoft.org/spalax/evariste',
        license="GPLv3 or any later version",
        #test_suite="jouets.test:suite", # TODO
        entry_points={
            'console_scripts': ['evariste = evariste.main:main']
            },
        classifiers=[ # TODO
            "Development Status :: 1 - Planning",
        ],
        long_description=readme(),
        zip_safe = False,
)



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>. 1

"""Directory representation and compilation."""

import configparser
import functools
import logging
import os
import pathlib
import tarfile

from evariste import action, plugins, utils

LOGGER = logging.getLogger(__name__)

@functools.total_ordering
class Tree:
    """Directory tree"""
    # pylint: disable=too-many-instance-attributes

    def __init__(self, path, parent=None, rcs=None, setup=None):
        self._subpath = {}
        self.report = None
        self.ignored = []
        self._config_files = []
        self.path = pathlib.Path(path)
        self.parent = parent
        self.config = None
        if parent is None:
            self.from_cwd = self.path
            self.from_source = pathlib.Path(".")
        else:
            self.from_cwd = parent.from_cwd / self.path
            self.from_source = parent.from_source / self.path
        if rcs is None:
            self.rcs = self.parent.rcs
        else:
            self.rcs = rcs
        if setup is None:
            self.setup = self.parent.setup
        else:
            self.setup = setup
        self._set_ignore()
        self._gather_config_files()

    def is_root(self):
        """Return ``True`` iff ``self`` is the root."""
        return self.parent is None

    @property
    def root(self):
        """Return the root of the tree."""
        if self.is_root():
            return self
        return self.parent.root

    def add_subpath(self, sub):
        """Add a subpath"""
        if len(sub.parts) > 1:
            self[sub.parts[0]].add_subpath(pathlib.Path(*sub.parts[1:]))
        else:
            self[sub.parts[0]] # pylint: disable=pointless-statement

    def __iter__(self):
        return iter(self._subpath)

    def keys(self):
        """Iterator over subpaths (as :class:`str` objects)."""
        yield from self._subpath.keys()

    def values(self):
        """Iterator over subpaths (as :class:`Tree` objects)."""
        yield from self._subpath.values()

    def __contains__(self, key):
        return key in self._subpath

    def __getitem__(self, key):
        if key not in self:
            if (self.from_cwd / key).is_dir():
                path_type = Directory
            else:
                path_type = File
            self._subpath[key] = path_type(key, parent=self)
        return self._subpath[key]

    def __delitem__(self, item):
        if item in self:
            del self._subpath[item]
            if len(self._subpath) == 0 and self.is_dir() and self.parent is not None:
                del self.parent[self.path.name]


    def __str__(self):
        return self.from_cwd.as_posix()

    def __len__(self):
        return len(self._subpath)

    def __eq__(self, other):
        return self.from_source == other.from_source

    def __lt__(self, other):
        return self.from_source < other.from_source

    def pprint(self, prefix="", last=True, print_function=print):
        """Pretty print of tree."""
        line = ""
        line += prefix
        if last:
            line += "└"
        else:
            line += "├"
        line += "─"
        if self.is_dir():
            line += "┬"
        else:
            line += "─"
        line += "╼ " + str(self.path)
        if last:
            postfix = "  "
        else:
            postfix = "│ "
        print_function(line)
        counter = 0
        for sub in sorted(self):
            counter += 1
            self[sub].pprint(
                prefix=prefix + postfix,
                last=(counter == len(self)),
                print_function=print_function
                )

    def is_dir(self):
        """Return `True` iff `self` is a directory."""
        return issubclass(self.__class__, Directory)

    def is_file(self):
        """Return `True` iff `self` is a file."""
        return issubclass(self.__class__, File)

    def walk(self, dirs=False, files=True):
        """Iterator over files or directories  of `self`.

        :param bool dirs: If `False`, do not yield directories.
        :param bool files: If `False`, do not yield files.
        """
        if (
                (dirs and self.is_dir())
                or
                (files and self.is_file())
            ):
            yield self
        for sub in sorted(self):
            yield from self[sub].walk(dirs, files)

    def prune_ignored(self):
        """Remove ignored paths from the tree."""
        # Recursive call
        for sub in self:
            self[sub].prune_ignored()

        # Process ``self``
        for regexp in self.ignored:
            for file in self.from_cwd.glob(regexp):
                self.root.prune(file.relative_to(self.root.from_cwd))

    def prune(self, path):
        """Remove a file.

        Argument can be either a :class:`pathlib.Path` or a :class:`tuple`.
        """
        if isinstance(path, tuple):
            parts = path
        elif isinstance(path, pathlib.Path):
            parts = path.parts
        else:
            raise ValueError
        if len(parts) == 1:
            del self[parts[0]]
        else:
            self[parts[0]].prune(parts[1:])

    @property
    def config_files(self):
        """Iterate over the names of configuration files affecting ``self``.

        This includes the configuration files of parent directories."""
        if self.parent is not None:
            yield from self.parent.config_files
        yield from self._config_files

    def set_config(self):
        """Set the configuration of current directory."""
        # Recursive call
        for sub in self:
            self[sub].set_config()

        # Set configuration of `self`
        self.config = configparser.ConfigParser(allow_no_value=True)
        for filename in self.config_files:
            with open(filename) as file:
                self.config.read_file(file)

    def _set_ignore(self):
        """Set the list of ignored regular expressions."""
        raise NotImplementedError()

    def _gather_config_files(self):
        """Iterator over the names of configuration files specific to ``self``
        """
        raise NotImplementedError()

    def compile_root(self, compilers):
        """Recursively compile files. This method is to be called by the root."""
        if not self.is_root():
            raise RuntimeError("Method 'compile_root()' can only be called on root of tree.")

        ## Attempt to multiprocess. Does not work because data is not shared.
        ## TODO: Implement multiprocess
        #pool = multiprocessing.pool.Pool()
        #self.compile(compilers, pool)
        #pool.close()
        #pool.join()
        self.compile(compilers)
        self.prune_depends()

    def compile(self, compilers, pool=None):
        """Recursively compile files."""
        raise NotImplementedError()

    def prune_depends(self):
        """Delete files that are dependencies of others."""
        raise NotImplementedError()

    def format(self, string):
        """Format given string, with several variables related to ``self.path``.
        """
        suffix = self.path.suffix
        if suffix.startswith("."):
            suffix = suffix[1:]
        return string.format(
            dirname=self.parent.from_cwd.as_posix(),
            filename=self.path.name,
            fullname=self.from_cwd.as_posix(),
            extension=suffix,
            basename=self.path.stem,
            )

    def render_readme(self, renderers, *args, **kwargs):
        """Find and render the readme corresponding to ``self``.

        If no readme was found, return an empty string.
        """
        for renderer in renderers.iter_plugins():
            for filename in renderer.iter_readme(
                    self.from_cwd,
                    directory=isinstance(self, Directory),
                ):
                return renderer.render(filename, *args, **kwargs)
        return ""

    def full_depends(self):
        """Return the list of all dependencies of this tree (recursively)."""
        raise NotImplementedError()


class Directory(Tree):
    """Directory"""

    def _set_ignore(self):
        """Set the list of ignored regular expressions."""
        ignorename = os.path.join(self.from_cwd.as_posix(), ".evsignore")
        if os.path.exists(ignorename):
            with open(ignorename) as file:
                self.ignored.extend([line.strip() for line in file.readlines()])

    def _gather_config_files(self):
        """Iterator over the names of configuration files specific to ``self``
        """
        name = os.path.join(self.from_cwd.as_posix(), ".evsconfig")
        if os.path.exists(name):
            self._config_files.append(name)

    def compile(self, compilers, pool=None):
        """Recursively compile files."""
        for sub in sorted(self):
            self[sub].compile(compilers, pool)
        self.report = action.DirectoryAction(self.setup).compile(self)

    def prune_depends(self):
        """Delete files that are dependencies of others."""
        # We cannot use `for sub in self:` here because dictionnary
        # `self._subpath` may change during iteration.
        for sub in list(self._subpath.keys()):
            if sub in self:
                self[sub].prune_depends()

    def full_depends(self):
        for sub in self:
            yield from self[sub].full_depends()


class File(Tree):
    """File"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.report = None

    def _set_ignore(self):
        """Set the list of ignored regular expressions."""
        for name in ["{}.ignore", ".{}.ignore"]:
            if os.path.exists(os.path.join(
                    self.parent.from_cwd.as_posix(),
                    name.format(self.path),
                )):
                self.parent.ignored.append(self.path.name)

    def _gather_config_files(self):
        """Set the configuration of current file."""
        for name in [".{}.evsconfig", "{}.evsconfig"]:
            name = os.path.join(
                self.parent.from_cwd.as_posix(),
                name.format(self.path),
                )
            if os.path.exists(name):
                self._config_files.append(name)

    def _compile(self, compile_function):
        """Actual compilation."""
        LOGGER.info("Compiling '{}'…".format(self.from_cwd))
        self.report = compile_function(self)
        if self.report.success:
            LOGGER.info("Compiling '{}': success.".format(self.from_cwd))
        else:
            LOGGER.info("Compiling '{}': failed.".format(self.from_cwd))

    def compile(self, compilers, pool=None):
        """Compile file, using multitasking poll if available."""
        try:
            if self.config.has_option('action', 'plugin'):
                compiler = compilers[self.config['action']['plugin']].plugin
            else:
                compiler = compilers.match(self.from_cwd)
        except plugins.NoMatch:
            del self.parent[self.path.name]
            return None

        if pool is None:
            self._compile(compiler.compile)
        else:
            pool.apply_async(self._compile, compiler.compile)


    def prune_depends(self):
        """Remove dependencies from tree."""
        if self.report:
            for path in self.report.depends:
                self.root.prune(
                    path.resolve().relative_to(
                        self.root.from_cwd.resolve()
                        )
                    )

    def make_archive(self, destdir):
        """Make an archive of ``self`` and its dependency.

        Stes are:
        - build the archive;
        - copy it to ``destdir``;
        - return the path of the archive, relative to ``destdir``.

        If ``self`` has no dependencies, consider the file as an archive.
        """
        if len(self.report.full_depends) == 1:
            utils.copy(self.from_cwd.as_posix(), (destdir / self.from_source).as_posix())
            return self.from_source
        else:
            archivepath = self.from_source.with_suffix(
                "{}.{}".format(self.from_source.suffix, 'tar.gz')
                )
            os.makedirs(os.path.dirname((destdir / archivepath).as_posix()), exist_ok=True)
            with tarfile.open(
                (destdir / archivepath).as_posix(),
                mode='w:gz',
                ) as archive:
                for file in self.report.full_depends:
                    archive.add(
                        file.as_posix(),
                        file.relative_to(self.parent.from_cwd).as_posix(),
                        )
            return archivepath

    def last_modified(self):
        """Return the last modified date and time of ``self``."""
        return self.rcs.last_modified(self.from_cwd)

    def full_depends(self):
        yield from self.report.full_depends


def build_tree(repository, setup):
    """Return a directory, fully set."""
    directory = repository.tree(setup)
    directory.set_config()
    directory.prune_ignored()
    return directory



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>. 1

"""Shell command: perform a (list of) shell command(s) on files."""

import logging
import os
import pathlib
import re
import subprocess
import threading

import evariste
from evariste import action

LOGGER = logging.getLogger(evariste.__name__)
STRACE_RE = re.compile(r'^open\("(?P<name>.*)",.*O_RDONLY.*\) = *[^ -].*')

def system(command, path, logreaders, depends):
    """Run a system command.

    This function:
    - run command;
    - log standard output and error;
    - track opened files.
    """

    def _process_strace_line(line):
        """Process output line of strace, and complete ``depends`` if relevant."""
        match = STRACE_RE.match(line)
        if match:
            try:
                name = pathlib.Path(
                    match.groupdict()["name"]
                    ).relative_to(
                        path.root.from_source
                    )
            except ValueError:
                return
            if name != path.path:
                if path.rcs.is_versionned(path.parent.from_cwd / name):
                    depends.add(path.parent.from_cwd / name)

    def _process_strace(pipe):
        """Process strace output, to find dependencies."""
        current = ""
        while 1:
            line = os.read(pipe, 1)
            if line == b'':
                break
            elif line == b'\n':
                _process_strace_line(current)
                current = ""
            else:
                current += line.decode()
        _process_strace_line(current)

    out3, in3 = os.pipe()
    process = subprocess.Popen(
        r"""strace -o "! sh -c 'cat >&{fd}'" -e trace=open {command}""".format(
            fd=in3,
            command=command,
            ),
        shell=True,
        stdin=subprocess.DEVNULL,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        pass_fds=(in3, out3),
        universal_newlines=True,
        cwd=path.parent.from_cwd.as_posix(),
        )

    threads = [
        threading.Thread(
            target=function,
            daemon=True,
            kwargs={
                'pipe' : pipe,
                }
            )
        for (function, pipe) in [
            (_process_strace, out3),
            (logreaders[0], process.stdout),
            (logreaders[1], process.stderr),
            ]
        ]
    for thread in threads:
        thread.start()
    process.wait()
    os.close(in3)
    for thread in threads:
        thread.join()
    os.close(out3)

    return process.returncode

class Command(action.Action):
    """Shell command action."""
    # pylint: disable=too-few-public-methods

    keyword = "command"

    def match(self, value): # pylint: disable=unused-argument
        """Return ``True`` iff ``value`` is to be compiled by ``self``."""
        return False

    def _compile(self, path):
        if not path.config.has_option(self.setup_section, 'command'):
            raise evariste.action.MissingOption(path.from_cwd, self.setup_section, 'command')
        command = path.format(path.config[self.setup_section]["command"])

        if not path.config.has_option(self.setup_section, 'target'):
            raise evariste.action.MissingOption(path.from_cwd, self.setup_section, 'target')

        LOGGER.log(15, "Runnig command: {}".format(command))

        pipes = [
            action.LogPipe(1),
            action.LogPipe(2),
            ]

        depends = set()
        returncode = system(
            command=command,
            path=path,
            logreaders=[pipe.read for pipe in pipes],
            depends=depends,
            )

        return action.Report(
            path,
            log=pipes,
            success=(returncode == 0),
            error="Command '{}' failed.".format(command),
            target=os.path.join(
                path.parent.from_source.as_posix(),
                path.format(path.config[self.setup_section]["target"])
                ),
            depends=depends,
            )

class MultiCommand(action.Action):
    """Chain of commands"""

    keyword = "multicommand"

    def __init__(self, setup):
        super().__init__(setup)
        self._commands = []
        self._report = None # Temporary report. It will be completed during compilation.

    def match(self, value):
        return False

    def iter_commands(self, env):
        """Iterator over the list of commands.

        :param object env: TODO
        """
        # pylint: disable=unused-argument
        yield from self._commands

    def run_subcommand(self, command):
        """Run a subcommand TODO"""
        LOGGER.log(15, "Runnig command: {}".format(command))
        pipes = [
            action.LogPipe(1),
            action.LogPipe(2),
            ]
        self._report.log.append(action.MultiLog(pipes))
        return 0 == system(
            command=command,
            path=self._report.path,
            logreaders=[log.read for log in pipes],
            depends=self._report.depends,
            )

    def _read_config(self, path, env=None):
        """Read and parse list of commands from configuration file.

        :param object env: TODO
        """
        # pylint: disable=unused-argument
        self._commands = []
        for option in path.config.options(self.setup_section):
            if option.startswith("cmd"):
                self._commands.append(
                    path.format(
                        path.config.get(self.setup_section, option)
                        )
                    )
        if not self._commands:
            raise action.MissingOption(path.from_cwd, self.setup_section, 'cmd*')
        if not path.config.has_option(self.setup_section, 'target'):
            raise action.MissingOption(path.from_cwd, self.setup_section, 'target')
        self._report.target = os.path.join(
            path.parent.from_source.as_posix(),
            path.format(path.config[self.setup_section]["target"])
            )


    def _compile(self, path, env=None):
        """TODO

        :param object env: TODO
        """
        # pylint: disable=arguments-differ
        self._report = action.Report(
            path,
            success=True,
            )
        self._read_config(path, env)
        for command in self.iter_commands(env):
            parsed_command = path.format(command)
            if not self.run_subcommand(parsed_command):
                self._report.success = False
                self._report.error = "Command '{}' failed.".format(parsed_command)
                break
        return self._report



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>. 1

"""Default HTML target renderer"""

from evariste.target.jinja2 import JinjaRenderer

class HtmlRenderer(JinjaRenderer):
    """Default HTML target renderer"""
    # pylint: disable=too-few-public-methods

    plugin_type = ["renderer", "html", "target"]

    keyword = "default"
    extension = "html"



# Copyright Louis Paternault 2015
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>. 1

"""Command line options"""

import argparse
import logging
import os
import textwrap

import evariste
from evariste import VERSION

LOGGER = logging.getLogger(evariste.__name__)

def _setup_type(name):
    """Check the argument and return its value.

    The argument must be an existing file.
    """
    if not os.path.exists(name):
        raise argparse.ArgumentTypeError(
            "File '{}' does not exist.".format(name)
            )
    if not os.path.isfile(name):
        raise argparse.ArgumentTypeError(
            "File '{}' is not a directory.".format(name)
            )
    return name

class Options:
    """Namespace of command line options."""

    _verbose = 1
    _logging_level = {
        0: 100, # Quiet
        1: 30, # Warning
        2: 20, # Info
        3: 15, # Details (custom level)
        4: 10, # Debug
        }

    def __str__(self):
        return str(dict([
            (attr, getattr(self, attr))
            for attr in dir(self)
            if not attr.startswith("_")
            ]))

    @property
    def verbose(self):
        """Get verbose level"""
        return self._verbose

    @verbose.setter
    def verbose(self, value):
        """Set verbose level (if not quiet)"""
        if self._verbose != 0:
            self._verbose = value
            LOGGER.setLevel(self._logging_level.get(self._verbose, 0))

    @property
    def quiet(self):
        """Is --quiet set?"""
        return self.verbose == 0

    @quiet.setter
    def quiet(self, dummy):
        """Set ---quiet"""
        self.verbose = 0

def commandline_parser():
    """Return a command line parser."""

    parser = argparse.ArgumentParser(
        description=(
            "Recursively compile files in a directory, and render result."
            ),
        formatter_class=argparse.RawTextHelpFormatter,
        )

    parser.add_argument(
        '--version',
        help='Show version',
        action='version',
        version='%(prog)s ' + VERSION
        )

    parser.add_argument(
        "-v", "--verbose",
        help='Verbose. Repeat for more details.',
        action='count',
        default=0,
        )

    parser.add_argument(
        "-q", "--quiet",
        help='Quiet. Does not print anything to standard output.',
        action='store_true',
        default=False,
        )

    parser.add_argument(
        'setup',
        metavar="SETUP",
        help=textwrap.dedent("""
            Setup file to process.
            """),
        type=_setup_type,
        )

    # TODO option -B --always-make
    # TODO option -n --dry-run
    # TODO option -r --render
    # TODO option -e --error stop on error
    # TODO option -j --jobs number of jobs

    return parser

def get_options():
    """Return the namespace of command line options."""
    return commandline_parser().parse_args(namespace=Options())
